#include <msp/core/raii.h>
#include <msp/gl/extensions/arb_direct_state_access.h>
#include <msp/gl/extensions/arb_texture_storage.h>
#include <msp/gl/extensions/msp_texture1d.h>
#include "bindable.h"
#include "error.h"
#include "texture1d.h"

using namespace std;

namespace Msp {
namespace GL {

Texture1D::Texture1D():
	Texture(GL_TEXTURE_1D),
	width(0),
	allocated(0)
{
	static Require _req(MSP_texture1D);
}

void Texture1D::storage(PixelFormat fmt, unsigned wd, unsigned lv)
{
	if(width>0)
		throw invalid_operation("Texture1D::storage");
	if(wd==0)
		throw invalid_argument("Texture1D::storage");

	set_format(fmt);
	width = wd;
	levels = get_n_levels();
	if(lv)
		levels = min(levels, lv);
}

void Texture1D::allocate(unsigned level)
{
	if(width==0)
		throw invalid_operation("Texture1D::allocate");
	if(level>=levels)
		throw invalid_argument("Texture1D::allocate");
	if(allocated&(1<<level))
		return;

	if(ARB_texture_storage)
	{
		Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
		if(ARB_direct_state_access)
			glTextureStorage1D(id, levels, storage_fmt, width);
		else
			glTexStorage1D(target, levels, storage_fmt, width);
		apply_swizzle();
		allocated |= (1<<levels)-1;
	}
	else
		image(level, 0);
}

void Texture1D::image(unsigned level, const void *data)
{
	if(width==0)
		throw invalid_operation("Texture1D::image");

	unsigned w = get_level_size(level);

	if(ARB_texture_storage)
		return sub_image(level, 0, w, data);

	BindRestore _bind(this);

	if(!allocated)
	{
		glTexParameteri(target, GL_TEXTURE_MAX_LEVEL, levels-1);
		apply_swizzle();
	}

	PixelComponents comp = get_components(storage_fmt);
	DataType type = get_component_type(storage_fmt);
	glTexImage1D(target, level, storage_fmt, w, 0, comp, type, data);

	allocated |= 1<<level;
	if(auto_gen_mipmap && level==0)
	{
		generate_mipmap();
		allocated |= (1<<levels)-1;
	}
}

void Texture1D::image(unsigned level, PixelComponents comp, DataType type, const void *data)
{
	if(comp!=get_components(format) || type!=get_component_type(format))
		throw incompatible_data("Texture1D::image");
	image(level, data);
}

void Texture1D::sub_image(unsigned level, int x, unsigned wd, const void *data)
{
	if(width==0)
		throw invalid_operation("Texture3D::image");

	Conditional<BindRestore> _bind(!ARB_direct_state_access, this);
	allocate(level);

	PixelComponents comp = get_components(storage_fmt);
	DataType type = get_component_type(storage_fmt);
	if(ARB_direct_state_access)
		glTextureSubImage1D(id, level, x, wd, comp, type, data);
	else
		glTexSubImage1D(target, level, x, wd, comp, type, data);

	if(auto_gen_mipmap && level==0)
		generate_mipmap();
}

void Texture1D::sub_image(unsigned level, int x, unsigned wd, PixelComponents comp, DataType type, const void *data)
{
	if(comp!=get_components(format) || type!=get_component_type(format))
		throw incompatible_data("Texture1D::sub_image");
	sub_image(level, x, wd, data);
}

void Texture1D::image(const Graphics::Image &img, unsigned lv)
{
	if(img.get_height()!=1)
		throw incompatible_data("Texture1D::image");

	unsigned w = img.get_width();
	PixelFormat fmt = pixelformat_from_image(img);
	if(width==0)
		storage(make_pixelformat(get_components(fmt), get_component_type(fmt), use_srgb_format), w, lv);
	else if(w!=width)
		throw incompatible_data("Texture1D::image");

	image(0, img.get_pixels());
}

unsigned Texture1D::get_n_levels() const
{
	unsigned n = 0;
	for(unsigned s=width; s; s>>=1, ++n) ;
	return n;
}

unsigned Texture1D::get_level_size(unsigned level) const
{
	return width>>level;
}

UInt64 Texture1D::get_data_size() const
{
	return id ? width*get_pixel_size(storage_fmt) : 0;
}


Texture1D::Loader::Loader(Texture1D &t):
	DataFile::DerivedObjectLoader<Texture1D, Texture::Loader>(t)
{
	init();
}

Texture1D::Loader::Loader(Texture1D &t, Collection &c):
	DataFile::DerivedObjectLoader<Texture1D, Texture::Loader>(t, c)
{
	init();
}

void Texture1D::Loader::init()
{
	add("raw_data", &Loader::raw_data);
	add("storage", &Loader::storage);
	add("storage", &Loader::storage_levels);
}

void Texture1D::Loader::raw_data(const string &data)
{
	obj.image(0, data.data());
}

void Texture1D::Loader::storage(PixelFormat fmt, unsigned w)
{
	obj.storage(fmt, w);
}

void Texture1D::Loader::storage_levels(PixelFormat fmt, unsigned w, unsigned l)
{
	obj.storage(fmt, w, l);
}

} // namespace GL
} // namespace Msp
